@@ -16,27 +16,151 @@
  *
  * */
 
+#include "debug.h"
 #include "iio-private.h"
 
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
 
-/* Include <winsock2.h> or <arpa/inet.h> for ntohl() */
-#ifdef _WIN32
-#include <winsock2.h>
-#else
-#include <arpa/inet.h>
-#endif
+static const char * const iio_chan_type_name_spec[] = {
+	[IIO_VOLTAGE] = "voltage",
+	[IIO_CURRENT] = "current",
+	[IIO_POWER] = "power",
+	[IIO_ACCEL] = "accel",
+	[IIO_ANGL_VEL] = "anglvel",
+	[IIO_MAGN] = "magn",
+	[IIO_LIGHT] = "illuminance",
+	[IIO_INTENSITY] = "intensity",
+	[IIO_PROXIMITY] = "proximity",
+	[IIO_TEMP] = "temp",
+	[IIO_INCLI] = "incli",
+	[IIO_ROT] = "rot",
+	[IIO_ANGL] = "angl",
+	[IIO_TIMESTAMP] = "timestamp",
+	[IIO_CAPACITANCE] = "capacitance",
+	[IIO_ALTVOLTAGE] = "altvoltage",
+	[IIO_CCT] = "cct",
+	[IIO_PRESSURE] = "pressure",
+	[IIO_HUMIDITYRELATIVE] = "humidityrelative",
+	[IIO_ACTIVITY] = "activity",
+	[IIO_STEPS] = "steps",
+	[IIO_ENERGY] = "energy",
+	[IIO_DISTANCE] = "distance",
+	[IIO_VELOCITY] = "velocity",
+	[IIO_CONCENTRATION] = "concentration",
+	[IIO_RESISTANCE] = "resistance",
+	[IIO_PH] = "ph",
+};
 
-/* winsock2.h defines ERROR, we don't want that */
-#undef ERROR
+static const char * const modifier_names[] = {
+	[IIO_MOD_X] = "x",
+	[IIO_MOD_Y] = "y",
+	[IIO_MOD_Z] = "z",
+	[IIO_MOD_X_AND_Y] = "x&y",
+	[IIO_MOD_X_AND_Z] = "x&z",
+	[IIO_MOD_Y_AND_Z] = "y&z",
+	[IIO_MOD_X_AND_Y_AND_Z] = "x&y&z",
+	[IIO_MOD_X_OR_Y] = "x|y",
+	[IIO_MOD_X_OR_Z] = "x|z",
+	[IIO_MOD_Y_OR_Z] = "y|z",
+	[IIO_MOD_X_OR_Y_OR_Z] = "x|y|z",
+	[IIO_MOD_ROOT_SUM_SQUARED_X_Y] = "sqrt(x^2+y^2)",
+	[IIO_MOD_SUM_SQUARED_X_Y_Z] = "x^2+y^2+z^2",
+	[IIO_MOD_LIGHT_BOTH] = "both",
+	[IIO_MOD_LIGHT_IR] = "ir",
+	[IIO_MOD_LIGHT_CLEAR] = "clear",
+	[IIO_MOD_LIGHT_RED] = "red",
+	[IIO_MOD_LIGHT_GREEN] = "green",
+	[IIO_MOD_LIGHT_BLUE] = "blue",
+	[IIO_MOD_QUATERNION] = "quaternion",
+	[IIO_MOD_TEMP_AMBIENT] = "ambient",
+	[IIO_MOD_TEMP_OBJECT] = "object",
+	[IIO_MOD_NORTH_MAGN] = "from_north_magnetic",
+	[IIO_MOD_NORTH_TRUE] = "from_north_true",
+	[IIO_MOD_NORTH_MAGN_TILT_COMP] = "from_north_magnetic_tilt_comp",
+	[IIO_MOD_NORTH_TRUE_TILT_COMP] = "from_north_true_tilt_comp",
+	[IIO_MOD_RUNNING] = "running",
+	[IIO_MOD_JOGGING] = "jogging",
+	[IIO_MOD_WALKING] = "walking",
+	[IIO_MOD_STILL] = "still",
+	[IIO_MOD_ROOT_SUM_SQUARED_X_Y_Z] = "sqrt(x^2+y^2+z^2)",
+	[IIO_MOD_I] = "i",
+	[IIO_MOD_Q] = "q",
+	[IIO_MOD_CO2] = "co2",
+	[IIO_MOD_VOC] = "voc",
+};
 
-#include "debug.h"
+/*
+ * Looks for a IIO channel modifier at the beginning of the string s. If a
+ * modifier was found the symbolic constant (IIO_MOD_*) is returned, otherwise
+ * IIO_NO_MOD is returned. If a modifier was found len_p will be updated with
+ * the length of the modifier.
+ */
+unsigned int find_channel_modifier(const char *s, size_t *len_p)
+{
+	unsigned int i;
+	size_t len;
 
-#ifndef _WIN32
-#define _strdup strdup
-#endif
+	for (i = 0; i < ARRAY_SIZE(modifier_names); i++) {
+		if (!modifier_names[i])
+			continue;
+		len = strlen(modifier_names[i]);
+		if (strncmp(s, modifier_names[i], len) == 0 &&
+				(s[len] == '\0' || s[len] == '_')) {
+			if (len_p)
+				*len_p = len;
+			return i;
+		}
+	}
+
+	return IIO_NO_MOD;
+}
+
+/*
+ * Initializes all auto-detected fields of the channel struct. Must be called
+ * after the channel has been otherwise fully initialized.
+ */
+void iio_channel_init_finalize(struct iio_channel *chn)
+{
+	unsigned int i;
+	size_t len;
+	char *mod;
+
+	chn->type = IIO_CHAN_TYPE_UNKNOWN;
+	chn->modifier = IIO_NO_MOD;
+
+	for (i = 0; i < ARRAY_SIZE(iio_chan_type_name_spec); i++) {
+		len = strlen(iio_chan_type_name_spec[i]);
+		if (strncmp(iio_chan_type_name_spec[i], chn->id, len) != 0)
+			continue;
+		/* Type must be followed by either a '_' or a digit */
+		if (chn->id[len] != '_' && chn->id[len] < '0' && chn->id[len] > '9')
+			continue;
+
+		chn->type = i;
+	}
+
+	mod = strchr(chn->id, '_');
+	if (!mod)
+		return;
+
+	mod++;
+
+	for (i = 0; i < ARRAY_SIZE(modifier_names); i++) {
+		if (!modifier_names[i])
+			continue;
+		len = strlen(modifier_names[i]);
+		if (strncmp(modifier_names[i], mod, len) != 0)
+			continue;
+		/* Modifier must be followed by a '_' */
+		if (mod[len] != '_')
+			continue;
+
+		chn->modifier = i;
+		break;
+	}
+}
 
 static char *get_attr_xml(struct iio_channel_attr *attr, size_t *length)
 {
@@ -182,6 +306,16 @@ bool iio_channel_is_scan_element(const struct iio_channel *chn)
 	return chn->is_scan_element;
 }
 
+enum iio_modifier iio_channel_get_modifier(const struct iio_channel *chn)
+{
+	return chn->modifier;
+}
+
+enum iio_chan_type iio_channel_get_type(const struct iio_channel *chn)
+{
+	return chn->type;
+}
+
 unsigned int iio_channel_get_attrs_count(const struct iio_channel *chn)
 {
 	return chn->nb_attrs;
@@ -602,7 +736,7 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 
 	for (i = 0; i < iio_channel_get_attrs_count(chn); i++) {
 		const char *attr = iio_channel_get_attr(chn, i);
-		int32_t len = (int32_t) ntohl(*(uint32_t *) ptr);
+		int32_t len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
 
 		ptr += 4;
 		if (len > 0) {
@@ -645,7 +779,7 @@ int iio_channel_attr_write_all(struct iio_channel *chn,
 		if (ret < 0)
 			goto err_free_buf;
 
-		*(int32_t *) ptr = (int32_t) htonl((uint32_t) ret);
+		*(int32_t *) ptr = (int32_t) iio_htobe32((uint32_t) ret);
 		ptr += 4;
 		len -= 4;
 
