@@ -16,11 +16,13 @@
  *
  * */
 
+#include <errno.h>
 #include <getopt.h>
 #include <iio.h>
 #include <signal.h>
 #include <stdio.h>
 #include <string.h>
+#include <errno.h>
 
 #define MY_NAME "iio_readdev"
 
@@ -30,28 +32,30 @@
 static const struct option options[] = {
 	  {"help", no_argument, 0, 'h'},
 	  {"network", required_argument, 0, 'n'},
-	  {"usb", required_argument, 0, 'u'},
+	  {"uri", required_argument, 0, 'u'},
 	  {"trigger", required_argument, 0, 't'},
 	  {"buffer-size", required_argument, 0, 'b'},
 	  {"samples", required_argument, 0, 's' },
+	  {"timeout", required_argument, 0, 'T'},
 	  {0, 0, 0, 0},
 };
 
 static const char *options_descriptions[] = {
 	"Show this help and quit.",
 	"Use the network backend with the provided hostname.",
-	"Use the USB backend with the device that matches the given VID/PID.",
+	"Use the context with the provided URI.",
 	"Use the specified trigger.",
 	"Size of the capture buffer. Default is 256.",
-	"Number of samples to capture, 0 = infinite. Default is 0."
+	"Number of samples to capture, 0 = infinite. Default is 0.",
+	"Buffer timeout in milliseconds. 0 = no timeout"
 };
 
 static void usage(void)
 {
 	unsigned int i;
 
-	printf("Usage:\n\t" MY_NAME " [-n <hostname>] [-u <vid>:<pid>] "
-			"[-t <trigger>] [-b <buffer-size>] [-s <samples>] "
+	printf("Usage:\n\t" MY_NAME " [-n <hostname>] [-t <trigger>] "
+			"[-b <buffer-size>] [-s <samples>] "
 			"<iio_device> [<channel> ...]\n\nOptions:\n");
 	for (i = 0; options[i].name; i++)
 		printf("\t-%c, --%s\n\t\t\t%s\n",
@@ -60,31 +64,124 @@ static void usage(void)
 }
 
 static struct iio_context *ctx;
-struct iio_buffer *buffer;
+static struct iio_buffer *buffer;
 static const char *trigger_name = NULL;
-unsigned int num_samples;
+static size_t num_samples;
 
-static bool app_running = true;
+static volatile sig_atomic_t app_running = true;
 static int exit_code = EXIT_SUCCESS;
 
 static void quit_all(int sig)
 {
 	exit_code = sig;
 	app_running = false;
+	if (buffer)
+		iio_buffer_cancel(buffer);
 }
 
-static void set_handler(int signal_nb, void (*handler)(int))
-{
 #ifdef _WIN32
-	signal(signal_nb, handler);
-#else
-	struct sigaction sig;
-	sigaction(signal_nb, NULL, &sig);
-	sig.sa_handler = handler;
-	sigaction(signal_nb, &sig, NULL);
-#endif
+
+#include <Windows.h>
+
+BOOL WINAPI sig_handler_fn(DWORD dwCtrlType)
+{
+	/* Runs in its own thread */
+
+	switch (dwCtrlType) {
+	case CTRL_C_EVENT:
+	case CTRL_CLOSE_EVENT:
+		quit_all(SIGTERM);
+		return TRUE;
+	default:
+		return FALSE;
+	}
 }
 
+static setup_sig_handler(void)
+{
+	SetConsoleCtrlHandler(sig_handler_fn, TRUE);
+}
+
+#elif NO_THREADS
+
+static void sig_handler(int sig)
+{
+	/*
+	 * If the main function is stuck waiting for data it will not abort. If the
+	 * user presses Ctrl+C a second time we abort without cleaning up.
+	 */
+	if (!app_running)
+		exit(sig);
+	app_running = false;
+}
+
+static void set_handler(int sig)
+{
+	struct sigaction action;
+
+	sigaction(sig, NULL, &action);
+	action.sa_handler = sig_handler;
+	sigaction(sig, &action, NULL);
+}
+
+static void setup_sig_handler(void)
+{
+	set_handler(SIGHUP);
+	set_handler(SIGPIPE);
+	set_handler(SIGINT);
+	set_handler(SIGSEGV);
+	set_handler(SIGTERM);
+}
+
+#else
+
+#include <pthread.h>
+
+static void * sig_handler_thd(void *data)
+{
+	sigset_t *mask = data;
+	int ret;
+
+	/* Blocks until one of the termination signals is received */
+	do {
+		ret = sigwaitinfo(mask, NULL);
+	} while (ret == -1 && errno == EINTR);
+
+	quit_all(ret);
+
+	return NULL;
+}
+
+static void setup_sig_handler(void)
+{
+	sigset_t mask, oldmask;
+	pthread_t thd;
+	int ret;
+
+	/*
+	 * Async signals are difficult to handle and the IIO API is not signal
+	 * safe. Use a seperate thread and handle the signals synchronous so we
+	 * can call iio_buffer_cancel().
+	 */
+
+	sigemptyset(&mask);
+	sigaddset(&mask, SIGHUP);
+	sigaddset(&mask, SIGPIPE);
+	sigaddset(&mask, SIGINT);
+	sigaddset(&mask, SIGSEGV);
+	sigaddset(&mask, SIGTERM);
+
+	pthread_sigmask(SIG_BLOCK, &mask, &oldmask);
+
+	ret = pthread_create(&thd, NULL, sig_handler_thd, &mask);
+	if (ret) {
+		fprintf(stderr, "Failed to create signal handler thread: %d\n", ret);
+		pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
+	}
+}
+
+#endif
+
 static struct iio_device * get_device(const struct iio_context *ctx,
 		const char *id)
 {
@@ -120,18 +217,19 @@ static ssize_t print_sample(const struct iio_channel *chn,
 			return -1;
 		}
 	}
-	return len;
+	return (ssize_t) len;
 }
 
 int main(int argc, char **argv)
 {
 	unsigned int i, nb_channels;
 	unsigned int buffer_size = SAMPLES_PER_READ;
-	int c, option_index = 0, arg_index = 0, ip_index = 0, device_index = 0;
+	int c, option_index = 0, arg_index = 0, ip_index = 0, uri_index = 0;
 	struct iio_device *dev;
 	size_t sample_size;
+	int timeout = -1;
 
-	while ((c = getopt_long(argc, argv, "+hn:u:t:b:s:",
+	while ((c = getopt_long(argc, argv, "+hn:u:t:b:s:T:",
 					options, &option_index)) != -1) {
 		switch (c) {
 		case 'h':
@@ -143,7 +241,7 @@ int main(int argc, char **argv)
 			break;
 		case 'u':
 			arg_index += 2;
-			device_index = arg_index;
+			uri_index = arg_index;
 			break;
 		case 't':
 			arg_index += 2;
@@ -157,6 +255,10 @@ int main(int argc, char **argv)
 			arg_index += 2;
 			num_samples = atoi(argv[arg_index]);
 			break;
+		case 'T':
+			arg_index += 2;
+			timeout = atoi(argv[arg_index]);
+			break;
 		case '?':
 			return EXIT_FAILURE;
 		}
@@ -168,43 +270,22 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
-	if (ip_index) {
+	setup_sig_handler();
+
+	if (uri_index)
+		ctx = iio_create_context_from_uri(argv[uri_index]);
+	else if (ip_index)
 		ctx = iio_create_network_context(argv[ip_index]);
-	} else if (device_index) {
-		char *ptr = argv[device_index], *end;
-		long vid, pid;
-
-		vid = strtol(ptr, &end, 0);
-		if (ptr == end || *end != ':') {
-			fprintf(stderr, "Invalid VID/PID\n");
-			return EXIT_FAILURE;
-		}
-
-		ptr = end + 1;
-		pid = strtol(ptr, &end, 0);
-		if (ptr == end) {
-			fprintf(stderr, "Invalid VID/PID\n");
-			return EXIT_FAILURE;
-		}
-
-		ctx = iio_create_usb_context(
-				(unsigned short) vid, (unsigned short) pid);
-	} else {
+	else
 		ctx = iio_create_default_context();
-	}
 
 	if (!ctx) {
 		fprintf(stderr, "Unable to create IIO context\n");
 		return EXIT_FAILURE;
 	}
 
-
-#ifndef _WIN32
-	set_handler(SIGHUP, &quit_all);
-#endif
-	set_handler(SIGINT, &quit_all);
-	set_handler(SIGSEGV, &quit_all);
-	set_handler(SIGTERM, &quit_all);
+	if (timeout >= 0)
+		iio_context_set_timeout(ctx, timeout);
 
 	dev = get_device(ctx, argv[arg_index + 1]);
 	if (!dev) {
@@ -247,7 +328,7 @@ int main(int argc, char **argv)
 		for (i = 0; i < nb_channels; i++) {
 			unsigned int j;
 			struct iio_channel *ch = iio_device_get_channel(dev, i);
-			for (j = arg_index + 2; j < argc; j++) {
+			for (j = arg_index + 2; j < (unsigned int) argc; j++) {
 				const char *n = iio_channel_get_name(ch);
 				if (!strcmp(argv[j], iio_channel_get_id(ch)) ||
 						(n && !strcmp(n, argv[j])))
@@ -260,7 +341,9 @@ int main(int argc, char **argv)
 
 	buffer = iio_device_create_buffer(dev, buffer_size, false);
 	if (!buffer) {
-		fprintf(stderr, "Unable to allocate buffer\n");
+		char buf[256];
+		iio_strerror(errno, buf, sizeof(buf));
+		fprintf(stderr, "Unable to allocate buffer: %s\n", buf);
 		iio_context_destroy(ctx);
 		return EXIT_FAILURE;
 	}
@@ -268,8 +351,11 @@ int main(int argc, char **argv)
 	while (app_running) {
 		int ret = iio_buffer_refill(buffer);
 		if (ret < 0) {
-			fprintf(stderr, "Unable to refill buffer: %s\n",
-					strerror(-ret));
+			if (app_running) {
+				char buf[256];
+				iio_strerror(-ret, buf, sizeof(buf));
+				fprintf(stderr, "Unable to refill buffer: %s\n", buf);
+			}
 			break;
 		}
 
@@ -277,19 +363,16 @@ int main(int argc, char **argv)
 		 * demux */
 		if (iio_buffer_step(buffer) == sample_size) {
 			void *start = iio_buffer_start(buffer);
-			ptrdiff_t len = (intptr_t) iio_buffer_end(buffer) -
-				(intptr_t) start;
-			size_t read_len;
+			size_t read_len, len = (intptr_t) iio_buffer_end(buffer)
+				- (intptr_t) start;
 
 			if (num_samples && len > num_samples * sample_size)
 				len = num_samples * sample_size;
 
 			for (read_len = len; len; ) {
-				ssize_t nb = fwrite(start, 1, len, stdout);
-				if (nb < 0) {
-					fprintf(stderr, "Unable to write data!\n");
+				size_t nb = fwrite(start, 1, len, stdout);
+				if (!nb)
 					goto err_destroy_buffer;
-				}
 
 				len -= nb;
 				start = (void *)((intptr_t) start + nb);
